from pathlib import Path

import numpy as np
from PIL import Image, ImageDraw

from skimage.io import imsave
from tifffile import imwrite

from model_server.accessors import GenericImageDataAccessor
from model_server.process import pad, rescale, resample_to_8bit

def _write_patch_to_file(where, fname, data):
    ext = fname.split('.')[-1].upper()
    where.mkdir(parents=True, exist_ok=True)

    if ext == 'PNG':
        assert data.dtype == 'uint8', f'Invalid data type {data.dtype}'
        assert data.shape[2] == 1, f'Cannot export multichannel images as PNGs; RGB not supported'
        assert data.shape[3] == 1, f'Cannot export z-stacks as PNGs'
        imsave(where / fname, data[:, :, 0, 0], check_contrast=False)
        return True

    elif ext in ['TIF', 'TIFF']:
        zcyx = np.moveaxis(data, [3, 2, 0, 1], [0, 1, 2, 3])
        imwrite(where / fname, zcyx, imagej=True)
        return True

    else:
        raise Exception(f'Unsupported file extension: {ext}')


def export_patches_from_zstack(
        where: Path,
        stack: GenericImageDataAccessor,
        zmask_meta: list,
        rescale_clip: float = 0.0,
        pad_to: int = 256,
        make_3d: bool = False,
        prefix='patch',
        **kwargs
):
    assert stack.chroma == 1, 'Expecting monochromatic image data'
    assert stack.nz > 1, 'Expecting z-stack'

    exported = []
    for mi in zmask_meta:
        obj = mi['info']
        sl = mi['slice']
        rbb = mi['relative_bounding_box']

        if make_3d:
            patch = stack.data[sl]
        else:
            patch = np.max(stack.data[sl], axis=3, keepdims=True)

        assert len(patch.shape) == 4
        assert patch.shape[2] == stack.chroma
        # assert patch.shape[3] == 1 # should not get to zstacks by this point

        if rescale_clip is not None:
            patch = rescale(patch, rescale_clip)

        if kwargs.get('draw_bounding_box') is True:
            x0 = rbb['x0']
            y0 = rbb['y0']
            x1 = rbb['x1']
            y1 = rbb['y1']

            for zi in range(0, patch.shape[3]):
                pilimg = Image.fromarray(patch[:, :, 0, zi])  # drawing modifies array in-place
                draw = ImageDraw.Draw(pilimg)
                draw.rectangle([(x0, y0), (x1, y1)], outline='white', width=kwargs.get('linewidth', 1))
                patch[:, :, 0, zi] = np.array(pilimg)

        if pad_to:
            patch = pad(patch, pad_to)

        ext = 'tif' if make_3d else 'png'
        fname = f'{prefix}-la{obj.label:04d}-zi{obj.zi:04d}.{ext}'
        success = _write_patch_to_file(where, fname, resample_to_8bit(patch))
        exported.append(fname)
    return success, exported